#include <MIDI.h>
#include <math.h>
#include "Mytypes.h"


// Создаем  MIDI интерфейс
MIDI_CREATE_DEFAULT_INSTANCE();

//Конфигурация пинов
#define CPU_D_0 2        //Шина данных bit 0 
#define CPU_D_1 4        //Шина данных bit 1
#define CPU_D_2 13       //Шина данных bit 2
#define CPU_D_3 14       //Шина данных bit 3
#define CPU_D_4 16       //Шина данных bit 4
#define CPU_D_5 17       //Шина данных bit 5
#define CPU_D_6 18       //Шина данных bit 6
#define CPU_D_7 19       //Шина данных bit 7
#define SH_AN0 34        //Аналоговый пин для педалей и регулятора громкости.
#define SPBLANK 33       //Out Enable для U3 74HCT374 (матрица светодиодов) в состоянии LOW пишем данные.
#define SPEN_L 32        //Out Enable для U8 74HC245 (комутатор управления устройствами) в состоянии LOW пишем данные.
#define CPU_WR_BUF_L 27  //Выбор направления передачи для U6 74HC245 (комутатор шины данных) в состоянии LOW пишем данные, в состоянии HIGH читаем данные
#define SCAN_CS_L 26     //Out Enable для U6 74HC245 (комутатор шины данных) в состоянии LOW пишем или читаем данные.
#define SPRES_L 25       //Выбор столбца: Вход элементов "И" для U5 74HC164 (клавиатура). Так же вход элементов "И" для U1 74HC164 (матрица светодиодов и кнопок).
#define SPREN_SW_L 23    //Out Enable для U4 74LVC245A в состоянии LOW читаем данные (матрица кнопок).
#define SPWEN_LED_L 22   //CLK для U1 74HC164 (матрица светодиодов и кнопок). Так же CLK для U3 74HCT374 (матрица светодиодов).
#define SPREN_KY_L 21    //Out Enable для U12 74HC245 в состоянии LOW читаем данные. Так же CLK для U5, U8, U9, LOW->HIGH - сдвигаем регистр. (клавиатура)

//пин записи в 74НС4051 (0-7) шина CPU_D_0 - CPU_D_4. Выбор педали или регулятора громкости.
//На шине 0 - Регклятор громкости.
//На шине 3 - Педаль Sustain
//На шине 2 - Педаль Sustenuto
//На шине 1 - Педаль CCPeda
#define SPEN_AN_L 15 //CLK для U16 74HCT374 (управление мультиплексором U15 74hc4051). Использует 5 младших бит шины данных.

#define AN_VOLUME 0
#define AN_SUSTENUTO 2
#define AN_SUSTAIN 3
#define AN_CCPEDA 1

//General MIDI контроллеры
#define Pedal_Sustain 64
#define Pedal_Sostenuto 66
#define Pedal_Soft 67
#define Control_Volume 7

#define DEBOUNCE_PERIOD 50000      // Период дребезга в микросекундах
#define STRONG_PRESS_PERIOD 7000   // период для самого сильного нажатия (2200 по умолчанию)
#define SOFT_PRESS_PERIOD 500000   // период для самого мягкого нажатия (500000 по умолчанию)

#define MIDI_CHANNEL 1             // Номер канала MIDI

uint8_t port_output = true;        //Состояние шины данных (запись/чтение)
keyp_t keys[88];
#define NUM_BANKS 11               //Количество банков (у нас 11 банков по 8 клавиш)

// Для сканирования банков
bank_t banks[NUM_BANKS];           //Текущие значения банков
bank_t prev_banks[NUM_BANKS];      //Предыдущие значения банков

void setup() {
  pinMode(CPU_D_0, OUTPUT);
  pinMode(CPU_D_1, OUTPUT);
  pinMode(CPU_D_2, OUTPUT);
  pinMode(CPU_D_3, OUTPUT);
  pinMode(CPU_D_4, OUTPUT);
  pinMode(CPU_D_5, OUTPUT);
  pinMode(CPU_D_6, OUTPUT);
  pinMode(CPU_D_7, OUTPUT);
	//Настройка портов управления на выход и уровень HIGH
  pinMode(SPBLANK,      OUTPUT);
  pinMode(SPEN_L,       OUTPUT);
  pinMode(CPU_WR_BUF_L, OUTPUT);
  pinMode(SCAN_CS_L,    OUTPUT);
  pinMode(SPRES_L,      OUTPUT);
  pinMode(SPREN_SW_L,   OUTPUT);
  pinMode(SPWEN_LED_L,  OUTPUT);
  pinMode(SPREN_KY_L,   OUTPUT);
  pinMode(SPEN_AN_L,    OUTPUT);
  digitalWrite(SPBLANK,      HIGH);
  digitalWrite(SPEN_L,       HIGH);
  digitalWrite(CPU_WR_BUF_L, HIGH);
  digitalWrite(SCAN_CS_L,    HIGH);
  digitalWrite(SPRES_L,      HIGH);
  digitalWrite(SPREN_SW_L,   HIGH);
  digitalWrite(SPWEN_LED_L,  HIGH);
  digitalWrite(SPREN_KY_L,   HIGH);
  digitalWrite(SPEN_AN_L,    HIGH);

  // Инициализация массива банков
  memset(banks, 0xff, sizeof(prev_banks));
  
  // Иницилизируем ноты с 21 для 88 клавиатуры
  for ( int key = 0; key < 88; key++) {
    keys[key].midi_note = 21 + key;    
    keys[key].t_up = 0;
    keys[key].t_down = 0;
    keys[key].t_pressed = 0;
    keys[key].t_released = 0;    
    keys[key].t_debounce = 0;
    
  }
  
  //Перед запуском следует отредактировать файл serialMIDI.h библиотеки MIDI Library v5.0.2
  //BaudRate установить на 115200 для использования Hairless MIDI
  //thruActivated установить в false так как функции приема MIDI сообщений не используем.
  MIDI.begin(MIDI_CHANNEL_OMNI);  // Прослушивать все входящие сообщения (оключил)
  delay(1000);
}


//чтение байта из шины
uint8_t read_port() {
  uint8_t data = 0;
  if (port_output){
    pinMode(CPU_D_0, INPUT);
    pinMode(CPU_D_1, INPUT);
    pinMode(CPU_D_2, INPUT);
    pinMode(CPU_D_3, INPUT);
    pinMode(CPU_D_4, INPUT);
    pinMode(CPU_D_5, INPUT);
    pinMode(CPU_D_6, INPUT);
    pinMode(CPU_D_7, INPUT);
    port_output = false;
  }

  data = digitalRead(CPU_D_7);
  data = data << 1;
  data = data | digitalRead(CPU_D_6);
  data = data << 1;
  data = data | digitalRead(CPU_D_5);
  data = data << 1;
  data = data | digitalRead(CPU_D_4);
  data = data << 1;
  data = data | digitalRead(CPU_D_3);
  data = data << 1;
  data = data | digitalRead(CPU_D_2);
  data = data << 1;
  data = data | digitalRead(CPU_D_1);
  data = data << 1;
  data = data | digitalRead(CPU_D_0);
  
  return data;
}

//звпись байта в шину
void write_port(uint8_t data) {  
  if(!port_output){
    pinMode(CPU_D_0, OUTPUT);
    pinMode(CPU_D_1, OUTPUT);
    pinMode(CPU_D_2, OUTPUT);
    pinMode(CPU_D_3, OUTPUT);
    pinMode(CPU_D_4, OUTPUT);
    pinMode(CPU_D_5, OUTPUT);
    pinMode(CPU_D_6, OUTPUT);
    pinMode(CPU_D_7, OUTPUT);
    port_output = true;
  }
  
  digitalWrite(CPU_D_7, data & 0b10000000);
  digitalWrite(CPU_D_6, data & 0b01000000);
  digitalWrite(CPU_D_5, data & 0b00100000);
  digitalWrite(CPU_D_4, data & 0b00010000);
  digitalWrite(CPU_D_3, data & 0b00001000);
  digitalWrite(CPU_D_2, data & 0b00000100);
  digitalWrite(CPU_D_1, data & 0b00000010);
  digitalWrite(CPU_D_0, data & 0b00000001);
}

//чтение аналоговых датчиков, на вход принимает номер сенсора 0-3
uint16_t read_analog_sensor(uint8_t sensor, bool await) {
  uint16_t value = 0;

  if(await){
    write_port(sensor);              //Пишем в шину данных номер входа для U15 74HC4051
    digitalWrite(CPU_WR_BUF_L, LOW); //Включаем U6 74HC245 на выход 
    digitalWrite(SCAN_CS_L, LOW);    //Включаем комутатор шины
    digitalWrite(SPEN_L, LOW);       //Включаем U8 74HC245 на выход  

    //Пишем в U16 данные для U15, просто дергаем CLK
    digitalWrite(SPEN_AN_L, LOW);  
    digitalWrite(SPEN_AN_L, HIGH);     
    
    //Конец чтения
    digitalWrite(SCAN_CS_L, HIGH);
    digitalWrite(CPU_WR_BUF_L, HIGH);  
    digitalWrite(SPEN_L, HIGH);

  }else{     
    digitalWrite(CPU_WR_BUF_L, LOW); //Включаем U6 74HC245 на выход 
    digitalWrite(SCAN_CS_L, LOW);    //Включаем комутатор шины
    digitalWrite(SPEN_L, LOW);       //Включаем U8 74HC245 на выход   

    value = analogRead(SH_AN0);      //Читаем состояние датчика   
    
    //Конец чтения
    digitalWrite(SCAN_CS_L, HIGH);
    digitalWrite(CPU_WR_BUF_L, HIGH);  
    digitalWrite(SPEN_L, HIGH);

    if(sensor){
      //Если это педаль то приводим к виду нажатая или нет
      if(value > 100){
        value = 0;
      }else{
        value = 127;
      }
    }
  } 

  return value;
}


//функция расчета скорости нажатия клавиши 
byte time_to_velocity(unsigned long time){
    float vel_t0 = STRONG_PRESS_PERIOD;  // период для самого сильного нажатия (2200 по умолчанию)
    float vel_t1 = SOFT_PRESS_PERIOD;    // период для самого мягкого нажатия (500000 по умолчанию)
    byte vel_pow = 7;                    // power law for mapping
    byte vel_min = 1;                    // наименьшее значение скорости для отправки
    byte vel = 0;                        // значение скорости для отправки

    float t = time - vel_t0;
    t = t < 0 ? 0:t;
    t /= vel_t1;
    t = t > 1 ? 1:t;
    t = (1.0-t);
    for(byte i = 1; i < vel_pow; i++){
      t *= t;
    }
    vel = (byte)(127.0 * t + 0.49);
    vel = vel > vel_min ? vel:vel_min;
    return vel;
  };



//Обрабатываем событие клавиш
void trigger(keyp_t *key, event_t event) {
  unsigned long t_value = 0;
  byte velocity = 0;

  if(event == KEY_PRESSED) {
    key->state = KEY_IS_GOING_DOWN;
    key->t_pressed = micros();
    
    
  } else if (event == KEY_DOWN) {
    key->state = KEY_IS_DOWN;    
    key->t_down = micros();
    t_value = key->t_down - key->t_pressed;
   
    //Если дребезг есть то выходим из функции
    if(key->t_down - key->t_debounce < DEBOUNCE_PERIOD){
      return;
    }

    velocity = time_to_velocity(t_value);
    MIDI.sendNoteOn(key->midi_note, velocity, MIDI_CHANNEL);
    
  } else if (event == KEY_RELEASED) {
    key->state = KEY_IS_GOING_UP;
    key->t_released = micros();
    key->t_debounce = micros();
    
  } else if ( event == KEY_UP) {
    key->state = KEY_IS_UP;
    key->t_up = micros();
    
    t_value = key->t_up - key->t_released;    
    velocity = time_to_velocity(t_value);
    MIDI.sendNoteOff(key->midi_note, velocity, MIDI_CHANNEL);    
    
  }
  
}

//Обрабатываем наджатие педалей и регулятор громкости
void footpedal() {  
  static uint8_t Sustain = 0;
  static uint8_t Sostenuto = 0;
  static uint8_t Soft = 0;
  static uint8_t Volume = 0;
  static uint8_t Volume_arr[16] = { 0 };
  static uint8_t count = 0;  

  static uint8_t prev_Sustain = 0;
  static uint8_t prev_Sostenuto = 0;
  static uint8_t prev_Soft = 0;
  static uint8_t prev_Volume = 0;

  static bool r_Sustain = false;
  static bool r_Sostenuto = false;
  static bool r_Soft = false;
  static bool r_Volume = false;

  static bool a_Sustain = true;
  static bool a_Sostenuto = false;
  static bool a_Soft = false;
  static bool a_Volume = false;

  static bool save_Volume = true;
  
  prev_Sustain = Sustain;
  prev_Sostenuto = Sostenuto;
  prev_Soft = Soft;  

  if(save_Volume){    
    prev_Volume = Volume;
    save_Volume = false;
  }

  if(a_Sustain){    
    read_analog_sensor(AN_SUSTAIN, a_Sustain);
    a_Sustain = false;
    r_Sustain = true;
  }

  if(a_Sostenuto){    
    read_analog_sensor(AN_SUSTENUTO, a_Sostenuto);
    a_Sostenuto = false;
    r_Sostenuto = true;
  }

  if(a_Soft){    
    read_analog_sensor(AN_CCPEDA, a_Soft);
    a_Soft = false;
    r_Soft = true;
  }

  if(a_Volume){    
    read_analog_sensor(AN_VOLUME, a_Volume);
    a_Volume = false;
    r_Volume = true;
  } 


  
  if (millis() - timer_await >= 5) { 
    timer_await = millis();

    if(r_Sustain){
      r_Sustain = false;
      Sustain = read_analog_sensor(AN_SUSTAIN, r_Sustain);
      a_Sostenuto = true;      
    }

    if(r_Sostenuto){
      r_Sostenuto = false;
      Sostenuto = read_analog_sensor(AN_SUSTENUTO, r_Sostenuto);      
      a_Soft = true;      
    }

    if(r_Soft){
      r_Soft = false;
      Soft = read_analog_sensor(AN_CCPEDA, r_Soft);
      a_Volume = true;      
    }

    if(r_Volume){
      r_Volume = false;            
      Volume_arr[count++] = map(read_analog_sensor(AN_VOLUME, r_Volume), 0, 4095, 0, 127);      
      a_Sustain = true;

      if(count == 16){                
        count = 0;        
        uint16_t v_s = 0;
        uint16_t v_max = 0;        
        for(int i = 0; i < 16; i++){          
          v_s = v_s + Volume_arr[i];
        }
        v_s = v_s >> 4;
        v_max = max(prev_Volume, Volume);
        v_s = v_s + v_max;
        v_s = v_s >> 1;
        Volume = v_s;
        
         
      }
            
    } 
  }

  if(Soft ^ prev_Soft){
    MIDI.sendControlChange(Pedal_Soft, Soft, MIDI_CHANNEL);
  }
  if(Sustain ^ prev_Sustain){
    MIDI.sendControlChange(Pedal_Sustain, Sustain, MIDI_CHANNEL);
  }
  if(Sostenuto ^ prev_Sostenuto){
    MIDI.sendControlChange(Pedal_Sostenuto, Sostenuto, MIDI_CHANNEL);
  }

  if(Volume ^ prev_Volume){
    MIDI.sendControlChange(Control_Volume, Volume, MIDI_CHANNEL);
    save_Volume = true;
  } 
}

//Выбор столбца для чтения, всего 22 столбца
void selects_row(uint8_t row) {   
  for(uint8_t i = 0; i < 24; i++){
    if(i == (24 - (row + 1))){
      digitalWrite(SPRES_L, LOW);
    }else{
      digitalWrite(SPRES_L, HIGH);      
    }
    //Дергаем CLK U9 74HC164 
    digitalWrite(SPREN_KY_L, LOW);
    digitalWrite(SPREN_KY_L, HIGH);
  }  
}

void scan() {

  digitalWrite(CPU_WR_BUF_L, HIGH);   //Включаем U6 74HC245 на вход
  digitalWrite(SPEN_L, LOW);          //Включаем U8 74HC245 на выход

  // Сканируем и сохраняем банки
  for(int bank = 0; bank < NUM_BANKS; bank++) {
    prev_banks[bank] = banks[bank];   // Сохраняем предыдущее состояние
    
    
    selects_row(bank * 2);            // Выбирает нижнюю строку BR    
    digitalWrite(SPREN_KY_L, LOW);    //Включаем U12 74HC245 на выход
    digitalWrite(SCAN_CS_L, LOW);     //Включаем комутатор шины
    banks[bank].bottom = read_port(); //Чтаем шину   
    digitalWrite(SCAN_CS_L, HIGH);    //Отключаем комутатор шины данных U6 74HC245
    digitalWrite(SPREN_KY_L, HIGH);
    
    
    selects_row(bank * 2 + 1);        // Выбирает верхнюю строку MK    
    digitalWrite(SPREN_KY_L, LOW);    //Включаем U12 74HC245 на выход
    digitalWrite(SCAN_CS_L, LOW);     //Включаем комутатор шины    
    banks[bank].top = read_port();    //Чтаем шину
    digitalWrite(SCAN_CS_L, HIGH);    //Отключаем комутатор шины данных U6 74HC245
    digitalWrite(SPREN_KY_L, HIGH);
  }
  digitalWrite(SPEN_L, HIGH);         //Отключаем комутатор устройств U8 74HC245
  
  // Обрабатываем события
  for(int bank = 0; bank < NUM_BANKS; bank++) {
    
    byte diff;
    
    
    diff = banks[bank].top ^ prev_banks[bank].top;
    if(diff) {
      for(int key = 0; key < 8; key++) {
        if(bitRead(diff, key)) { 
          event_t event = bitRead(banks[bank].top, key) ? KEY_UP : KEY_PRESSED;
          trigger(&keys[bank * 8 + key], event);
        }
      }
    }
    
    
    diff = banks[bank].bottom ^ prev_banks[bank].bottom;
    if(diff) {
      for(int key = 0; key < 8; key++) {
        if(bitRead(diff, key)) { 
          event_t event = bitRead(banks[bank].bottom, key) ? KEY_RELEASED : KEY_DOWN;           
          trigger(&keys[bank * 8 + key], event);
        }
      }
    }
    
  }

}

void loop() {
  scan();
  footpedal();
}